import readline
import sys
import tomllib
from pathlib import Path

from rich.console import Console
from rich.markdown import Markdown

from .api import DuckChat
from .exceptions import DuckChatException
from .models import ModelType

HELP_MSG = (
    "\033[1;1m- /help         \033[0mDisplay the help message\n"
    "\033[1;1m- /singleline   \033[0mEnable singleline mode, validate is done by <enter>\n"
    "\033[1;1m- /multiline    \033[0mEnable multiline mode, validate is done by EOF <Ctrl+D>\n"
    "\033[1;1m- /quit         \033[0mQuit\n"
    "\033[1;1m- /retry        \033[0mRegenerate answer to № prompt (default /retry 1)"
)

COMMANDS = {"help", "singleline", "multiline", "quit", "retry"}


def completer(text: str, state: int) -> str | None:
    origline = readline.get_line_buffer()
    words = origline.split()
    if not origline.startswith("/"):
        return None
    if len(words) < 2 and words[0][1:] not in COMMANDS:
        options = [cmd for cmd in COMMANDS if cmd.startswith(text)]
        if state < len(options):
            return options[state]
    return None


class CLI:

    def __init__(self) -> None:
        readline.parse_and_bind("tab: complete")
        readline.set_completer(completer)
        self.INPUT_MODE = "singleline"
        self.COUNT = 1
        self.console = Console()

    async def run(self) -> None:
        """Base loop program"""
        model = self.read_model_from_conf()
        print(f"Using \033[1;4m{model.value}\033[0m")
        async with DuckChat(model) as chat:
            print("Type \033[1;4m/help\033[0m to display the help")

            while True:
                print(f"\033[1;4m>>> User input №{self.COUNT}:\033[0m", end="\n")

                user_input = self.get_user_input()

                # user input is command
                if user_input.startswith("/"):
                    await self.command_parsing(user_input.split(), chat)
                    continue

                # empty user input
                if not user_input:
                    print("Bad input")
                    continue

                print(f"\033[1;4m>>> Response №{self.COUNT}:\033[0m", end="\n")
                try:
                    self.answer_print(await chat.ask_question(user_input))
                except DuckChatException as e:
                    print(f"Error occurred: {str(e)}")
                else:
                    self.COUNT += 1

    def get_user_input(self) -> str:
        if self.INPUT_MODE == "singleline":
            try:
                user_input = input()
            except EOFError:
                return ""
        else:
            contents = []
            while True:
                try:
                    line = input()
                except EOFError:
                    break
                contents.append(line)
            user_input = "".join(contents)
        return user_input.strip()

    def switch_input_mode(self, mode: str) -> None:
        if mode == "singleline":
            self.INPUT_MODE = "singleline"
            print("Switched to singleline mode, validate is done by <enter>")
        else:
            self.INPUT_MODE = "multiline"
            print("Switched to multiline mode, validate is done by EOF <Ctrl+D>")

    async def command_parsing(self, args: list[str], chat: DuckChat) -> None:
        """Recognize command"""
        print("\033[1;4m>>> Command response:\033[0m")
        match args[0][1:]:
            case "singleline":
                self.switch_input_mode("singleline")
            case "multiline":
                self.switch_input_mode("multiline")
            case "quit":
                print("Quit")
                sys.exit(0)
            case "help":
                print(HELP_MSG)
            case "retry":
                if self.COUNT == 1:
                    return
                try:
                    count = int(args[1])
                except Exception:
                    count = len(chat.vqd) - 1
                if count < 0:
                    count = -count
                if count >= len(chat.vqd):
                    count = len(chat.vqd) - 1
                print(f"\033[1;4m>>> REDO Response №{count}:\033[0m", end="\n")
                try:
                    self.answer_print(await chat.reask_question(count))
                except DuckChatException as e:
                    print(f"Error occurred: {str(e)}")
                else:
                    self.COUNT = count + 1
            case _:
                print("Command doesn't find")
                print("Type \033[1;4m/help\033[0m to display the help")

    def answer_print(self, query: str) -> None:
        if "`" in query:  # block of code
            self.console.print(Markdown(query))
        else:
            print(query)

    def read_model_from_conf(self) -> ModelType:
        filepath = Path.home() / ".config" / "hey" / "conf.toml"
        if filepath.exists():
            with open(filepath, "rb") as f:
                conf = tomllib.load(f)
                model_name = conf["model"]
            if model_name in (x.name for x in ModelType):
                if model_name == "GPT3":
                    print("\033[1;1m GPT3 is deprecated! Use GPT4o\033[0m")
                return ModelType[model_name]
        return ModelType.Claude


def safe_entry_point() -> None:
    import asyncio

    asyncio.run(CLI().run())



import argparse
import glob
import os
import subprocess
import sys
from itertools import chain
from pathlib import Path

try:
    import black  # noqa
    import isort  # noqa
    import ruff  # noqa
except Exception as e:
    print(e)
    print("Error: did you run env and install full requirements?")
    quit()

BASE_DIR = Path().parent


def find_dirs() -> list[str]:
    """Get all dirs, not in .gitignore"""
    exclude_dirs = []
    gitignore = BASE_DIR / ".gitignore"
    if gitignore.exists():
        with open(gitignore, "r") as f:
            exclude_dirs = f.readlines()
        exclude_dirs = [x.strip().replace("/", "") for x in exclude_dirs]
        exclude_dirs.append(".git")
    dirs = []
    for f in BASE_DIR.iterdir():
        if f.is_dir() and f.name not in exclude_dirs:
            dirs.append(f)
    return dirs


def find_files(dirs: list) -> list[Path]:
    """Get all python files recursive"""
    files = glob.iglob("*.py", root_dir=BASE_DIR)
    for _dir in dirs:
        f = glob.iglob(f"{_dir}/**/*.py", root_dir=BASE_DIR, recursive=True)
        files = chain(files, f)

    _files = []
    for file in files:
        _file = BASE_DIR.absolute() / file
        _files.append(_file)
    return _files


def parse_args():
    def FilePath(astring):
        file = Path(astring)
        if not file.exists():
            raise argparse.ArgumentTypeError(f"File={file} does not exists!")
        return file

    parser = argparse.ArgumentParser(
        description="Code formatter with ruff, isort and black.",
    )
    parser.add_argument(
        "--files",
        metavar="file",
        type=FilePath,
        nargs="+",
        help="path to file",
        required=False,
    )
    parser.add_argument(
        "--fix",
        action="store",
        nargs="*",
        help="fix code with ruff",
        required=False,
    )
    parser.add_argument(
        "--silent",
        action="store",
        nargs="*",
        help="silent mode, log only errors",
        required=False,
    )

    args = parser.parse_args()
    return args


def run():
    args = parse_args()
    print("Formatting code")

    stdout = None
    if args.silent is not None:
        stdout = subprocess.DEVNULL
        sys.stdout = open(os.devnull, "w")

    files = args.files
    if not files:
        print("Finding files")
        dirs = find_dirs()
        files = find_files(dirs)

    if args.fix is None:
        print("Ruff fixing")
        subprocess.run(
            ["ruff", "check", *files, "--unsafe-fixes", "--fix"], stdout=stdout
        )

    print("Isort fix imports")
    subprocess.run(["isort", *files], stdout=stdout)

    print("Black linting")
    subprocess.run(["black", *files], stderr=stdout, stdout=stdout)

    print("---------Done---------")

    print("Errors unfixed:")
    subprocess.run(["ruff", "check", *files])


if __name__ == "__main__":
    run()



from enum import Enum

import msgspec


class ModelType(Enum):
    Claude = "claude-3-haiku-20240307"
    GPT3 = "gpt-3.5-turbo-0125"  # Deprecated
    Llama = "meta-llama/Llama-3-70b-chat-hf"
    GPT4o = "gpt-4o-mini"
    Mixtral = "mistralai/Mixtral-8x7B-Instruct-v0.1"


class Role(Enum):
    user = "user"
    assistant = "assistant"


class Message(msgspec.Struct):
    role: Role
    content: str


class History(msgspec.Struct):
    model: ModelType
    messages: list[Message]

    def add_input(self, message: str) -> None:
        self.messages.append(Message(Role.user, message))

    def add_answer(self, message: str) -> None:
        self.messages.append(Message(Role.assistant, message))



from types import TracebackType
from typing import Self

import aiohttp
import msgspec
from fake_useragent import UserAgent

from .exceptions import (
    ConversationLimitException,
    DuckChatException,
    RatelimitException,
)
from .models import History, ModelType


class DuckChat:
    def __init__(
        self,
        model: ModelType = ModelType.Claude,
        session: aiohttp.ClientSession | None = None,
        user_agent: UserAgent | str = UserAgent(min_version=120.0),
    ) -> None:
        if type(user_agent) is str:
            self.user_agent = user_agent
        else:
            self.user_agent = user_agent.random  # type: ignore

        self._session = session or aiohttp.ClientSession(
            headers={
                "Host": "duckduckgo.com",
                "Accept": "text/event-stream",
                "Accept-Language": "en-US,en;q=0.5",
                "Accept-Encoding": "gzip, deflate, br",
                "Referer": "https://duckduckgo.com/",
                "User-Agent": self.user_agent,
                "DNT": "1",
                "Sec-GPC": "1",
                "Connection": "keep-alive",
                "Sec-Fetch-Dest": "empty",
                "Sec-Fetch-Mode": "cors",
                "Sec-Fetch-Site": "same-origin",
                "TE": "trailers",
            }
        )
        self.vqd: list[str | None] = []
        self.history = History(model, [])
        self.__encoder = msgspec.json.Encoder()
        self.__decoder = msgspec.json.Decoder()

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        await self._session.__aexit__(exc_type, exc_value, traceback)

    async def get_vqd(self) -> None:
        """Get new x-vqd-4 token"""
        async with self._session.get(
            "https://duckduckgo.com/duckchat/v1/status", headers={"x-vqd-accept": "1"}
        ) as response:
            if response.status == 429:
                res = await response.read()
                try:
                    err_message = self.__decoder.decode(res).get("type", "")
                except Exception:
                    raise DuckChatException(res.decode())
                else:
                    raise RatelimitException(err_message)
            self.vqd.append(response.headers.get("x-vqd-4"))
            if not self.vqd:
                raise DuckChatException("No x-vqd-4")

    async def get_answer(self) -> str:
        """Get message answer from chatbot"""
        async with self._session.post(
            "https://duckduckgo.com/duckchat/v1/chat",
            headers={
                "Content-Type": "application/json",
                "x-vqd-4": self.vqd[-1],
            },
            data=self.__encoder.encode(self.history),
        ) as response:
            res = await response.read()
            if response.status == 429:
                raise RatelimitException(res.decode())
            try:
                data = self.__decoder.decode(
                    b"["
                    + b",".join(
                        res.lstrip(b"data: ")
                        .rstrip(b"\n\ndata: [DONE][LIMIT_CONVERSATION]\n")
                        .split(b"\n\ndata: ")
                    )
                    + b"]"
                )
            except Exception:
                raise DuckChatException(f"Couldn't parse body={res.decode()}")
            message = []
            for x in data:
                if x.get("action") == "error":
                    err_message = x.get("type", "") or str(x)
                    if x.get("status") == 429:
                        if err_message == "ERR_CONVERSATION_LIMIT":
                            raise ConversationLimitException(err_message)
                        raise RatelimitException(err_message)
                    raise DuckChatException(err_message)
                message.append(x.get("message", ""))
        self.vqd.append(response.headers.get("x-vqd-4"))
        return "".join(message)

    async def ask_question(self, query: str) -> str:
        """Get answer from chat AI"""
        if not self.vqd:
            await self.get_vqd()
        self.history.add_input(query)

        message = await self.get_answer()

        self.history.add_answer(message)
        return message

    async def reask_question(self, num: int) -> str:
        """Get answer from chat AI"""

        if num >= len(self.vqd):
            num = len(self.vqd) - 1
        self.vqd = self.vqd[:num]

        if not self.history.messages:
            return ""

        if not self.vqd:
            await self.get_vqd()
            self.history.messages = [self.history.messages[0]]
        else:
            num = min(num, len(self.vqd))
            self.history.messages = self.history.messages[: (num * 2 - 1)]
        message = await self.get_answer()
        self.history.add_answer(message)

        return message



import aiohttp


class DuckChatException(aiohttp.client.ClientError):
    """Base exception class for duck_chat."""


class RatelimitException(DuckChatException):
    """Raised for rate limit exceeded errors during API requests."""


class ConversationLimitException(DuckChatException):
    """Raised for conversation limit during API requests to AI endpoint."""
